"use strict";

function BUFFER (size, init_value) {
  this.size = size;
  this.array = new Array(size);
  this.array.fill(init_value);
  this.cursor = 0;
};

BUFFER.prototype = {
  set: function (index, value) {
    this.array[index] = value;
  },

  get: function (index) {
    return this.array[index];
  },

  add_cursor: function (value) {
    this.cursor = this.cursor + value;
  },

  allocate: function (size) {
    let return_address = this.cursor;
    this.add_cursor(size);
    return return_address;
  }
};

function STACK (size, init_value) {
  BUFFER.call(this, size, init_value);
};

STACK.prototype = Object.create(BUFFER.prototype);

STACK.prototype.push = function (value) {
  this.set(this.cursor, value);
  this.add_cursor(1);
};

STACK.prototype.pop = function () {
  this.add_cursor(-1);
  return this.get(this.cursor);
};

function HASH_TABLE_ENTRY (index) {
  this.index = index;
  this.key = null;
  this.value = null;
  this.orbit_length = 0;
  this.orbiton = 0;
};

HASH_TABLE_ENTRY.prototype = {
  occured: function () {
    return this.key !== null;
  },

  used: function () {
    return this.value !== null;
  },

  no_collision: function () {
    return this.index === this.orbiton;
  }
};

function HASH_TABLE (size,
                     key_equal,
                     hash) {
  this.size = size;
  this.key_equal = key_equal;
  this.hash = hash;
  this.array = new Array(size);
  this.counter = 0;
  for (let i = 0; i < size ; i++) {
    this.array[i] = new HASH_TABLE_ENTRY(i);
  };
};

HASH_TABLE.prototype = {
  insert: function (key) {
    // key -> index
    //     -> null -- denotes the hash_table is filled
    let orbit_index = this.hash(key, 0);
    let counter = 0;
    while (true) {
      let index = this.hash(key, counter);
      let entry = this.index_to_entry(index);
      if (!entry.occured()) {
        entry.key = key;
        entry.orbiton = orbit_index;
        let orbit_entry = this.index_to_entry(orbit_index);
        orbit_entry.orbit_length = 1 + counter;
        this.counter = 1 + this.counter;
        return index;
      } else if (this.key_equal(key, entry.key)) {
        return index;
      } else if (counter === this.size) {
        return null;
      } else {
        counter = 1 + counter;
      }
    }
  },

  search: function (key) {
    // key -> index
    //     -> null -- denotes key not occured
    let counter = 0;
    while (true) {
      let index = this.hash(key, counter);
      let entry = this.index_to_entry(index);
      if (!entry.occured()) {
        return null;
      } else if (this.key_equal(key, entry.key)) {
        return index;
      } else if (counter === this.size) {
        return null;
      } else {
        counter = 1 + counter;
      }
    }
  },

  key_to_index: function (key) {
    let index = this.insert(key);
    if (index !== null) {
      return index;
    } else {
      console.log("hash_table is filled");
      throw "hash_table is filled";
    }
  },

  index_to_entry: function (index) {
    return this.array[index];
  },

  key_to_entry: function (key) {
    return index_to_entry(key_to_index(key));
  },

  report_orbit: function (index, counter) {
    let entry = this.index_to_entry(index);
    while (true) {
      if (counter >= entry.orbit_length) {
        return;
      } else {
        let key = entry.key;
        let next_index = this.hash(key, counter);
        let next_entry = this.index_to_entry(next_index);
        if (index === next_entry.orbiton) {
          console.log("  - {", next_index, "} ",
                      next_entry.key); }
        counter = 1 + counter;
      }
    }
  },

  report: function () {
    console.log("\n");
    console.log("- hash_table-table report_used");
    let index = 0;
    while (true) {
      let entry = this.index_to_entry(index);
      if (index === this.size) {
        break;
      } else {
        if (entry.occured() &&
            entry.no_collision()) {
          console.log("  - {", index, "} ",
                      entry.key, " # ",
                      entry.orbit_length);
          if (entry.used()) {
            console.log("      ", entry.value);
          }
          this.report_orbit(index, 1);
        }
        index = 1 + index;
      }
    }
    console.log("\n");
    console.log("- used : ", this.counter);
    console.log("- free : ", this.size - this.counter);
  }
};

let cell_area = new BUFFER(1024 * 1024, 0);

let argument_stack = new STACK(1024, 0);
let return_stack = new STACK(1024, 0);

let string_to_sum = (string) => {
  let sum = 0;
  let max_step = 10;
  for (let i = 0; i < string.length; i++) {
    sum = sum +
      string.codePointAt(i) *
      (2 << Math.min(i, max_step));
  }
  return sum;
};

let tag_hash_table = new HASH_TABLE(
  // prime table size
  997,
  // key_equal
  (key1, key2) => {
    return key1 === key2;
  },
  // hash
  (key, counter) => {
    return (counter + string_to_sum(key)) % 997;
  }
);

// tag_hash_table.key_to_index("123");

function array_equal (a, b) {
  if (a === b) return true;
  if (a == null || b == null) return false;
  if (a.length != b.length) return false;
  for (var i = 0; i < a.length; ++i) {
    if (a[i] !== b[i]) return false;
  }
  return true;
}

Array.prototype.last = function () {
  return this[this.length - 1];
};

function array_remove_dup (orig_array) {
  let array = orig_array.map((element) => {return element;});
  array.sort();
  let result = [];
  for (let i = 0; i < array.length; i++) {
    if (result.last() === array[i]) {
    } else {
      result.push(array[i]);
    }
  }
  return result;
}

/*
  tag_hash_table.key_to_index("testkey1");
  tag_hash_table.key_to_index("testkey2");
  tag_hash_table.key_to_index("testkey3");
  tag_hash_table.report();
*/

let tagset_hash_table = new HASH_TABLE(
  // prime table size
  997,
  // key_equal
  (key1, key2) => {
    return array_equal(array_remove_dup(key1),
                       array_remove_dup(key2));
  },
  // hash
  (key, counter) => {
    let sum =
        array_remove_dup(key)
        .map(string_to_sum)
        .reduce((sum, v) => {return sum + v;});
    return (counter + sum) % 997;
  }
);

/*
  tagset_hash_table.key_to_index(["testkey1"]);
  tagset_hash_table.key_to_index(["testkey2"]);
  tagset_hash_table.key_to_index(["testkey3"]);
  tagset_hash_table.key_to_index(["testkey4"]);
  tagset_hash_table.key_to_index(["testkey1", "testkey2", "testkey3"]);
  tagset_hash_table.key_to_index(["testkey1", "testkey2"]);
  tagset_hash_table.key_to_index(["testkey4", "testkey4", "testkey1"]);
  tagset_hash_table.key_to_index(["testkey4", "testkey1"]);
  tagset_hash_table.report();
*/

let primitive_function_record = new BUFFER(1024, 0);

let create_primitive_function = (fun) => {
  let return_address = primitive_function_record.cursor;
  primitive_function_record
    .set(return_address, fun);
  primitive_function_record.add_cursor(1);
  return return_address;
};

var address_after_explainer = 0;

let interpreter = () => {
  try {
    while (true) {
      let jojo = return_stack.pop();
      let jo = cell_area.get(jojo);
      let explainer = cell_area.get(jo);
      return_stack.push(jojo + 1);
      address_after_explainer = jo + 1;
      primitive_function_record.get(explainer).call();
      // continue;
    }
  } catch (string) {
    switch (string) {
    case "bye":
      break;
    }
  }
};

let in_host_tag_record = new Map();

let data = (value) => {
  cell_area.set(cell_area.cursor, value);
  cell_area.add_cursor(1);
};

let mark = (tag_string) => {
    in_host_tag_record.set(tag_string, cell_area.cursor);
};

let link = 0;

let define_header = (tag_string, explainer) => {
  data(link);
  link = cell_area.cursor - 1;
  mark(tag_string);
  data(explainer);
};

let define_primitive_function = (tag_string, fun) => {
  let function_index = create_primitive_function(fun);
  define_header(tag_string, primitive_function_explainer);
  data(function_index);
};

let define_function = (tag_string, function_tag_string_array) => {
  define_header(tag_string, function_explainer);
  function_tag_string_array.forEach(
    function_tag_string => {
      data(in_host_tag_record
           .get(function_tag_string));
    }
  );
};

let define_variable = (tag_string, value) => {
  define_header(tag_string, variable_explainer);
  data(value);
};

let primitive_function_explainer = create_primitive_function(
  () => {
    primitive_function_record.get(
      cell_area.get(address_after_explainer)
    ).call();
  });

let function_explainer = create_primitive_function(
  () => {
    return_stack.push(address_after_explainer);
  });

let variable_explainer = create_primitive_function(
  () => {
    argument_stack.push(
      cell_area.get(address_after_explainer));
  });

define_primitive_function(
  "end",
  () => {
    return_stack.pop();
  }
);

define_primitive_function(
  "bye",
  () => {
    console.log("bye bye ^-^/");
    throw "bye";
  }
);

define_primitive_function(
  "dup",
  () => {
    let a = argument_stack.pop();
    argument_stack.push(a);
    argument_stack.push(a);
  }
);

define_primitive_function(
  "mul",
  () => {
    let a = argument_stack.pop();
    let b = argument_stack.pop();
    argument_stack.push(a * b);
  }
);

define_primitive_function(
  "simple-wirte",
  () => {
    console.log(argument_stack.pop());
  }
);

define_variable("little-test-number", 4);

define_function(
  "square",
  [ "dup",
    "mul",
    "end"
  ]
);

define_function(
  "little-test",
  [ "little-test-number",
    "square",
    "simple-wirte",
    "bye"
  ]
);

define_function(
  "first-function",
  [ "little-test",
    "end"
  ]
);

let begin_to_interpret_threaded_code = () => {
  let function_body_for_little_test =
        in_host_tag_record.get("first-function") + 1;
  return_stack.push(function_body_for_little_test);
  interpreter();
};

begin_to_interpret_threaded_code();
